import docker
import json
import os

from .MLILException import MLILException


class ModelManager:
    '''
    ModelManager class used to abstract away the deployment of MLflow models to dedicated containers
    '''

    def __init__(
        self,
        model_image: str = 'ghcr.io/mlinsightlab/mlinsightlab-model-container:latest',
        model_network: str = 'mlinsightlab_model_network',
        mlflow_tracking_uri: str = 'http://mlflow:2244',
        model_port: str = '8888'
    ):
        '''
        Parameters
        ----------
        model_image : str (default 'ghcr.io/mlinsightlab/mlinsightlab-model-container:latest)'
            The image of the container to use
        model_network : str (default 'mlinsightlab_model_network')
            The network to deploy the models to
        mlflow_tracking_uri : str (default 'http://mlflow:2244')
            The tracking URI for the MLflow service on the docker network
        model_port : str (default '8888')
            The port to deploy the model to on the container
        '''

        # Create docker container client
        self.docker_client = docker.from_env()

        # Store the image, network, and port on which the images will be stored
        self.model_image = model_image if model_image else os.environ['MODEL_CONTAINER_IMAGE']
        self.model_network = model_network if model_network else os.environ['MODEL_NETWORK']
        self.mlflow_tracking_uri = mlflow_tracking_uri if mlflow_tracking_uri else os.environ[
            'MLFLOW_TRACKING_URI']
        self.container_port = model_port if model_port else os.environ['MODEL_PORT']

        # Store the containers
        self.models = []

    def deploy_model(
            self,
            model_uri: str,
            model_name: str,
            model_flavor: str,
            model_version_or_alias: str,
            use_gpu: bool = False,
            volumes: dict = None,
            requirements: str = None,
            kwargs: dict = None,
            quantization_kwargs: dict = None
    ):
        '''
        Deploy a containerized model

        Parameters
        ----------
        model_uri : str
            The URI of the model according to MLflow
        model_name : str
            The name of the model
        model_flavor : str
            The flavor of the model
        model_version_or_alias : str
            The version or alias of the model
        use_gpu : bool (default False)
            If true, will allow the container access to available GPUs
        volumes : dict or None (default None)
            If provided, a dictionary of volumes to mount to the container
        requirements: str or None (default None)
            Additional pip requirements needed to deploy the model
        kwargs: dict or None (default None)
            Additional keyword arguments needed to deploy the model
        quantization_kwargs: dict or None (default None)
            Additional quantization keyword arguments needed to deploy the model

        Returns
        -------
        success : bool
            Returns True if successful
        '''

        # Environment variables for the contianer
        environment = {
            'MODEL_URI': model_uri,
            'MODEL_FLAVOR': model_flavor,
            'MLFLOW_TRACKING_URI': self.mlflow_tracking_uri
        }
        if requirements:
            environment['REQUIREMENTS'] = requirements
        if kwargs:
            environment['KWARGS'] = json.dumps(kwargs)
        if quantization_kwargs:
            environment['QUANTIZATION_KWARGS'] = json.dumps(
                quantization_kwargs)

        # Name for the container
        container_name = f'mlinsightlab__model__{model_name}__{model_flavor}__{model_version_or_alias}'

        # Run the container, giving it access to the GPU if requested
        if use_gpu:
            model_container = self.docker_client.containers.run(
                self.model_image,
                auto_remove=True,
                environment=environment,
                network=self.model_network,
                name=container_name,
                detach=True,
                device_requests=[
                    docker.types.DeviceRequest(
                        count=-1, capabilities=[['gpu']])
                ],
                volumes=volumes
            )
        else:
            model_container = self.docker_client.containers.run(
                self.model_image,
                auto_remove=True,
                environment=environment,
                network=self.model_network,
                name=container_name,
                detach=True,
                volumes=volumes
            )

        # Append the container properties to the models list
        self.models.append(
            {
                'model_name': model_name,
                'model_flavor': model_flavor,
                'model_version_or_alias': model_version_or_alias,
                'container_name': model_container.name
            }
        )

        return True

    def remove_deployed_model(
            self,
            model_name: str,
            model_flavor: str,
            model_version_or_alias: str,
    ):
        '''
        Remove a deployed model

        Parameters
        ----------
        model_name : str
            The name of the model
        model_flavor : str
            The flavor of the model
        model_version_or_alias : str
            The version or alias of the model

        Returns
        -------
        success : bool
            Returns True if successful
        '''

        # Search for the container name
        container_name = None
        for model in self.models:
            if model['model_name'] == model_name and model['model_flavor'] == model_flavor and model['model_version_or_alias'] == model_version_or_alias:
                container_name = model['container_name']
                break

        # Raise exception if container not found
        if container_name is None:
            raise MLILException('Container for that model not found')

        # Get the container, raise exception if not found
        try:
            container = self.docker_client.containers.get(container_name)
        except Exception:
            raise MLILException('Container for that model not found')

        # Try to stop the container, raise exception if unable to
        try:
            container.stop()
            self.models = [
                model for model in self.models if model['container_name'] != container_name
            ]
        except Exception as e:
            raise MLILException(
                f'Error trying to stop containerized model: {str(e)}')

        return True

    def remove_all_models(self):
        '''
        Remove all models
        '''

        # Go through the models and remove them all
        for model in self.models:
            self.remove_deployed_model(
                model['model_name'],
                model['model_flavor'],
                model['model_version_or_alias']
            )

        return True

    def get_model_status(
            self,
            model_name: str,
            model_flavor: str,
            model_version_or_alias: str
    ):
        '''
        Get the status of a deployed model

        Parameters
        ----------
        model_name : str
            The name of the model
        model_flavor : str
            The flavor of the model
        model_version_or_alias : str
            The version or alias of the model

        Returns
        -------
        status : str
            The status of the model container
        '''

        # Search for the container name
        container_name = None
        for model in self.models:
            if model['model_name'] == model_name and model['model_flavor'] == model_flavor and model['model_version_or_alias'] == model_version_or_alias:
                container_name = model['container_name']
                break

        # Raise exception if not found
        if not container_name:
            raise MLILException('Container for that model not found')

        # Return status
        return self.docker_client.containers.get(container_name).status

    def get_model_logs(
            self,
            model_name: str,
            model_flavor: str,
            model_version_or_alias: str
    ):
        '''
        Get the logs of a deployed model

        Parameters
        ----------
        model_name : str
            The name of the model
        model_flavor : str
            The flavor of the model
        model_version_or_alias : str
            The version or alias of the model

        Returns
        -------
        logs : str
            The logs of the model container
        '''

        # Search for the container name
        container_name = None
        for model in self.models:
            if model['model_name'] == model_name and model['model_flavor'] == model_flavor and model['model_version_or_alias'] == model_version_or_alias:
                container_name = model['container_name']
                break

        # Raise exception if not found
        if not container_name:
            raise MLILException('Container for that model not found')

        # Return logs
        return self.docker_client.containers.get(container_name).logs().decode('utf-8')



"""
Utilities for interacting with the Machine Learning Insights Lab (MLIL) platform.

This package is based around the 'MLIL_client' class, which has the capabilities to
interact with the functionality of the MLIL platform from a Python CLI environment.
"""

from .ModelManager import ModelManager
from .MLILClient import MLILClient

__all__ = ['MLILClient', 'ModelManager']
__version__ = '0.0.27'



# Helper functions to manage and interat with models in the platform
from .MLILException import MLILException
from typing import Union, List, Optional
from .endpoints import DEPLOY_MODEL_ENDPOINT, LIST_MODELS_ENDPOINT, UNDEPLOY_MODEL_ENDPOINT, PREDICT_ENDPOINT, GET_MODEL_LOGS
import pandas as pd
import requests


def _deploy_model(
    url: str,
    creds: dict,
    model_name: str,
    model_flavor: str,
    model_version_or_alias: str,
    load_request: dict,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Deploys a model.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds: dict
        Dictionary that must contain keys 'username' and 'key', and associated values.
    model_name: str
        The name of the model to load
    model_flavor: str
        The flavor of the model, e.g. 'transformers', 'pyfunc', etc.
    model_version_or_alias: str
        The version of the model that you wish to load (from MLFlow).
    load_request : dict = None
        A dictionary containing additional parameters for loading the model,
        including 'requirements', 'quantization_kwargs', and 'kwargs'.
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    if load_request is None:
        load_request = {
            'model_name': model_name,
            'model_flavor': model_flavor,
            'model_version_or_alias': model_version_or_alias
        }
    else:
        load_request['model_name'] = model_name
        load_request['model_flavor'] = model_flavor
        load_request['model_version_or_alias'] = model_version_or_alias

    # Format the URL
    url = f'''{
        url}/{DEPLOY_MODEL_ENDPOINT}'''

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.post(
            url,
            auth=(creds['username'], creds['key']),
            json=load_request,
            verify=ssl_verify
        )

    # If request is not successful, raise appropriate exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _list_models(
    url: str,
    creds: dict,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Lists all *deployed* models. To view undeployed models, check the MLFlow UI.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the URL
    url = f'{url}/{LIST_MODELS_ENDPOINT}'

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.get(
            url,
            auth=(creds['username'], creds['key']),
            verify=ssl_verify
        )

    # If the request is not successful, raise an appropriate exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _undeploy_model(
    url: str,
    creds: dict,
    model_name: str,
    model_flavor: str,
    model_version_or_alias: str,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Unloads a model from the deployment service.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds: dict
        Dictionary that must contain keys 'username' and 'key', and associated values.
    model_name: str
        The name of the model to unload.
    model_flavor: str
        The flavor of the model, e.g. 'transformers', 'pyfunc', etc.
    model_version_or_alias: str
        The version of the model that you wish to unload (from MLFlow).
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the URL
    url = f'''{
        url}/{UNDEPLOY_MODEL_ENDPOINT}/{model_name}/{model_flavor}/{model_version_or_alias}'''

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.delete(
            url,
            auth=(creds['username'], creds['key']),
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _predict(
    url: str,
    creds: dict,
    model_name: str,
    model_flavor: str,
    model_version_or_alias: str,
    data: Union[list, int, float, str],
    convert_to_numpy: bool = True,
    predict_function: str = 'predict',
    dtype: str = None,
    params: Optional[dict] = None,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Calls the 'predict' function of the specified deployed model.

    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    model_name: str
        The name of the model to be invoked.
    model_flavor: str
        The flavor of the model, e.g. 'transformers', 'pyfunc', etc.
    model_version_or_alias: str
        The version of the model that you wish to invoke (from MLFlow).
    data: varied types
        The input data for prediction. Can be a single value or list of values
    convert_to_numpy: bool = True
        Whether or not to convert inputs to a NumPy array.
    predict_function: str, optional
        The name of the prediction function to call. Default is 'predict'.
    dtype: str, optional
        The data type of the input
    params: dict, optional
        Additional parameters for the prediction.
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the data so that it is a list
    if isinstance(data, str):
        data = [data]

    # Format the JSON payload
    json_data = {
        'model_name': model_name,
        'model_flavor': model_flavor,
        'model_version_or_alias': model_version_or_alias,
        'data': data,
        'predict_function': predict_function,
        'params': params if params else {},
        'convert_to_numpy': convert_to_numpy
    }

    # Add dtype to the JSON payload, if provided
    if dtype:
        json_data.update({'dtype': dtype})

    # Format the URL
    url = f'''{url}/{PREDICT_ENDPOINT}'''

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.post(
            url,
            auth=(creds['username'], creds['key']),
            json=json_data,
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _get_model_logs(
    url: str,
    creds: dict,
    model_name: str,
    model_flavor: str,
    model_version_or_alias: str,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Gets model logs.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment to the platform.
    creds: dict
        Dictionary that must contain keys 'username' and 'key', and associated values.
    model_name: str
        The name of the model to unload.
    model_flavor: str
        The flavor of the model, e.g. 'transformers', 'pyfunc', etc.
    model_version_or_alias: str
        The version of the model that you wish to unload (from MLFlow).
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the URL
    url = f'''{
        url}/{GET_MODEL_LOGS}/{model_name}/{model_flavor}/{model_version_or_alias}'''

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.get(
            url,
            auth=(creds['username'], creds['key']),
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp



# Manage users and credentials
from .MLILException import MLILException
from .endpoints import CREATE_USER_ENDPOINT, DELETE_USER_ENDPOINT, VERIFY_PASSWORD_ENDPOINT, NEW_PASSWORD_ENDPOINT, GET_USER_ROLE_ENDPOINT, UPDATE_USER_ROLE_ENDPOINT, LIST_USERS_ENDPOINT
import requests


def _create_user(
    url: str,
    creds: dict,
    username: str,
    role: str,
    api_key: str | None,
    password: str | None,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Create a user within the platform.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    cred:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    username: str
        The user's display name and login credential
    role: str
        The role to be given to the user
    api_key: str or NULL
        An API key for the new user
    password: str or NULL
        Password for user login
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the URL
    url = f'{url}/{CREATE_USER_ENDPOINT}'

    # Format the JSON payload
    json_data = {
        'username': username,
        'role': role
    }

    # Format additional parameters for API key and password, if provided
    if api_key:
        json_data['api_key'] = api_key
    if password:
        json_data['password'] = password

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.post(
            url,
            auth=(creds['username'], creds['key']),
            json=json_data,
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _delete_user(
    url: str,
    creds: dict,
    username: str,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Delete a user within the platform.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    username: str
        The user's display name and login credential
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    json_data = {
        'username': username
    }

    url = f'{url}/{DELETE_USER_ENDPOINT}/{username}'

    with requests.Session() as sess:
        resp = sess.delete(
            url,
            auth=(creds['username'], creds['key']),
            json=json_data,
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _verify_password(
    url: str,
    creds: dict,
    username: str,
    password: str,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Verify a user's password.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    username: str
        The user's display name and login credential
    password: str
        Password for user login
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    json_data = {
        'username': username,
        'password': password
    }

    url = f'{url}/{VERIFY_PASSWORD_ENDPOINT}'

    with requests.Session() as sess:
        resp = sess.post(
            url,
            auth=(creds['username'], creds['key']),
            json=json_data,
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _issue_new_password(
    url: str,
    creds: dict,
    username: str,
    new_password: str,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Create a new a password for an existing user.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    username: str
        The user's display name and login credential
    new_password: str
        New password for user authentication
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the URL
    url = f'{url}/{NEW_PASSWORD_ENDPOINT}/{username}'

    # Format the JSON payload
    json_data = {
        'username': username,
        'new_password': new_password
    }

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.put(
            url,
            auth=(creds['username'], creds['key']),
            json=json_data,
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(resp.json())
    return resp


def _get_user_role(
    url: str,
    creds: dict,
    username: str,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Check a user's role.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    username: str
        The user's display name and login credential.
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the URL
    url = f'{url}/{GET_USER_ROLE_ENDPOINT}/{username}'

    # Format the JSON payload
    json_data = {
        'username': username
    }

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.get(
            url,
            auth=(creds['username'], creds['key']),
            json=json_data,
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _update_user_role(
    url: str,
    creds: dict,
    username: str,
    new_role: str,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Update a user's role.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    username: str
        The user's display name and login credential
    new_role: str
        New role to attribute to the specified user
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the URL
    url = f'{url}/{UPDATE_USER_ROLE_ENDPOINT}/{username}'

    # Format the JSON payload
    json_data = {
        'username': username,
        'new_role': new_role
    }

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.put(
            url,
            auth=(creds['username'], creds['key']),
            json=json_data,
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp


def _list_users(
    url: str,
    creds: dict,
    ssl_verify: bool = True
):
    '''
    NOT MEANT TO BE CALLED BY THE END USER

    Update a user's role.
    Called within the MLILClient class.

    Parameters
    ----------
    url: str
        String containing the URL of your deployment of the platform.
    creds:
        Dictionary that must contain keys 'username' and 'key', and associated values.
    ssl_verify: bool (default True)
        Whether to verify SSL certificates in the request
    '''

    # Format the URL
    url = f'{url}/{LIST_USERS_ENDPOINT}'

    # Make the request to the platform
    with requests.Session() as sess:
        resp = sess.get(
            url,
            auth=(creds['username'], creds['key']),
            verify=ssl_verify
        )

    # If the request is not successful, raise exception, else return response
    if not resp.ok:
        raise MLILException(str(resp.json()))
    return resp
